/*
 * Copyright (C) 2010 Andrew Kozin AKA Stanson <stanson@pangolin.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dsp_open(), dsp_play() and dsp_close() functions for ALSA
 */

#include <alsa/asoundlib.h>
#include "log.h"
#include "audio.h"

#define SND_DEVICE	"plughw:0,0"	/* default sound device */
#define MIN_FRAMES	2

static snd_pcm_t	*pcm = NULL;
static snd_output_t	*snd_log = NULL;

int dsp_open( char * name, int verbose )
{
    int ret, dir = 0;
    snd_pcm_hw_params_t *hw_params;
    snd_pcm_sw_params_t *sw_params;
    snd_pcm_uframes_t	buffer_size = PCM_SAMPLES_IN_FRAME * 50;
    snd_pcm_uframes_t	period_size = PCM_SAMPLES_IN_FRAME;

    snd_output_stdio_attach(&snd_log, stderr, 0);

    ret = snd_pcm_open( &pcm, name[0] ? name : SND_DEVICE, SND_PCM_STREAM_PLAYBACK, 0 );
    if( ret < 0 )
    {
        _logf( LOG_ERR, "Playback open error: %s\n", snd_strerror( ret ) );
        return ret;
    }

    ret = snd_pcm_hw_params_malloc(&hw_params);
    if( ret < 0 )
    {
        _logf( LOG_ERR, "Can't alloc hardware struct %s\n", snd_strerror( ret ) );
	goto out_err;
    }

    snd_pcm_hw_params_any( pcm, hw_params );
    snd_pcm_hw_params_set_rate_resample( pcm, hw_params, 1 );
    snd_pcm_hw_params_set_access( pcm, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED );
    snd_pcm_hw_params_set_format( pcm, hw_params, SND_PCM_FORMAT_S16_LE );
    snd_pcm_hw_params_set_rate( pcm, hw_params, 8000, 0 );
    snd_pcm_hw_params_set_channels( pcm, hw_params, 1 );
    snd_pcm_hw_params_set_buffer_size_near( pcm, hw_params, &buffer_size );
    snd_pcm_hw_params_set_period_size_near( pcm, hw_params, &period_size, &dir );

    ret = snd_pcm_hw_params( pcm, hw_params );
    if( ret < 0 )
    {
	_logf( LOG_ERR, "Can't set hardware parameters %s\n", snd_strerror( ret ) );
        goto out_err;
    }

    ret = snd_pcm_sw_params_malloc( &sw_params );
    if( ret < 0 )
    {
        _logf( LOG_ERR, "Can't alloc software struct %s\n", snd_strerror( ret ) );
        goto out_err;
    }

    snd_pcm_sw_params_current( pcm, sw_params );
    snd_pcm_sw_params_set_start_threshold( pcm, sw_params, period_size * MIN_FRAMES );
    snd_pcm_sw_params_set_avail_min( pcm, sw_params, period_size);
    snd_pcm_sw_params_set_silence_threshold( pcm, sw_params, period_size / 2 );

    ret = snd_pcm_sw_params( pcm, sw_params );
    if( ret < 0 )
    {
	_logf( LOG_ERR, "Can't set software parameters: %s\n", snd_strerror(ret) );
	goto out_err;
    }

    if( verbose ) snd_pcm_dump( pcm, snd_log );

    /* initially stop playback */
    snd_pcm_drop( pcm );
    snd_pcm_prepare( pcm );
    return 0;

out_err:
    snd_pcm_close( pcm );
    return ret;
}

int dsp_play( short * data, int len )
{
    int ret;
    len /= PCM_SAMPLE_SIZE;

    while( len )
    {
	ret = snd_pcm_writei( pcm, data, len );
	if( ret < 0 )
	{
	    if( ret == -EAGAIN ) continue;
	    if( ret == -EPIPE )
	    {
		ret = snd_pcm_prepare( pcm );
		if( ret < 0 )
		{
		    _logf( LOG_ERR, "PCM recovery error: %s\n", snd_strerror(ret) );
		}
	    }
	    else
	    {
		_logf( LOG_ERR, "PCM write error: %s\n", snd_strerror(ret) );
	    }
	    break;
	}
	else if( ret > 0 )
	{
	    data += ret;
	    len -= ret;
	}
    }
    return 0;
}

void dsp_close( void ) { if( pcm ) snd_pcm_close( pcm ); }
