/*
 * Copyright (C) 2010 Andrew Kozin AKA Stanson <stanson@pangolin.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#define __USE_GNU
#include <poll.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <pthread.h>

#include "conf.h"
#include "log.h"
#include "sdp.h"
#include "rfcomm.h"
#include "sco.h"
#include "hci.h"
#include "audio.h"

#include "../config.h"

struct thread_data {
    int			fd;
    char		*raddr;
};

char		g_device[PATH_MAX] = { 0 };
char		g_btdev[256] = { 0 };
char		g_phone[256] = { 0 };
int		g_verbose = 0;
int		g_daemon = 1;
int		g_syslog = 0;
int		g_persist = 0;

int		g_sco_threads = 0;

/* Bluetooth globals */
sdp_session_t	*g_sdp_hfp = 0;
int		g_channel = -1;
char		g_service_desc[256] = { 0 };
char		g_service_prov[256] = { 0 };

/* RFCOMM data */
int		g_rfcommfd = -1;

/* SCO data */
int		g_scofd = -1;

struct _conf_params	params[] =
{
    /*  name         type    variable             default value   */
    { "device",       's', { &g_device },       "plughw:0,0" },
    { "btdev",        's', { &g_btdev },        "hci0" },
    { "phone",        's', { &g_phone },        "" },
    { "service_prov", 's', { &g_service_prov }, "Mini HandsFree" },
    { "service_desc", 's', { &g_service_desc }, "Mini HandsFree Profile daemon" },
    { "verbose",      'b', { &g_verbose },      "no" },
    { "daemon",       'b', { &g_daemon },       "yes" },
    { "syslog",       'b', { &g_syslog },       "no" },
    { "persist",      'b', { &g_persist },      "no" },
    { NULL, 0, { NULL }, NULL }
};

int daemon_start(void)
{
  pid_t  cpid;

  cpid = fork();
  if(cpid < 0)
  {
    fprintf(stderr, "Cannot fork to become a daemon!\n");
    return 0;
  }
  else if(cpid > 0) /* parent exits */
  {
    exit(0);
  }
  setsid();
  close(STDIN_FILENO);
  close(STDOUT_FILENO);
  close(STDERR_FILENO);
  return 1;
}

void sig_handler( int sig )
{
    switch( sig )
    {
	case SIGINT:
	    _logf( LOG_NOTICE, "Got signal INT, exiting.\n" );
	    break;
	case SIGTERM:
	    _logf( LOG_NOTICE, "Got signal TERM, exiting.\n" );
	    break;
	default:
	    return;
    }
    _logf( LOG_INFO, "Shutting down SCO server.\n" );
    close( g_scofd );
    _logf( LOG_INFO, "Closing RFCOMM connection.\n" );
    close( g_rfcommfd );
    _logf( LOG_INFO, "Deregistering HandsFree Profile.\n" );
    sdp_deregister_hfp( g_sdp_hfp );
    if( g_syslog ) closelog();
    dsp_close();
    exit( 0 );
}

int slc_established()
{
    _logf( LOG_INFO, "Established Sevice Level Connection.\n" );
//    _logf( LOG_INFO, "Establishing SCO connection...\n" );
//    g_scofd = sco_do_connect( g_phone, 48 );
    return 0;
}

void * handle_connection( void * arg )
{
    struct thread_data	*td = (struct thread_data *)arg;
    unsigned char	buf[2048];
    struct pollfd	fds;
    int			ret;

    if( !td ) { pthread_exit( NULL ); return NULL; }

    g_sco_threads++;
    _logf( LOG_INFO, "SCO threads: %d\n", g_sco_threads );

    fds.fd = td->fd;
    fds.events = POLLIN;
    fds.revents = 0;

    while( 1 )
    {
	ret = poll( &fds, 1, -1 );
	if( ret < 0 )
	{
	    _logf( LOG_ERR, "SCO thread %s poll() error: %s\n", td->raddr, strerror( errno ) );
	    break;
	}

	if( fds.revents & ( POLLERR | POLLNVAL | POLLRDHUP ) )
	{ // SCO connection lost
	    _logf( LOG_INFO, "SCO thread %s disconnected\n", td->raddr );
	    break;
	}

	if( fds.revents & POLLHUP )
	{ // hangup
	    _logf( LOG_ERR, "SCO thread %s hang up\n", td->raddr );
	    break;
	}

	if( fds.revents & POLLIN )
	{ /* Receive */
	    ret = recv( td->fd, buf, sizeof(buf), 0 );
	    if( ret > 0 )
	    {
		dsp_play( (short *) buf, ret );
	    }
	    else if ( ret == 0 )
	    {
		_logf( LOG_ERR, "SCO thread %s peer closed connection\n", td->raddr );
		break;
	    }
	    else
	    { // read error
		_logf( LOG_ERR, "SCO thread %s read error\n", td->raddr );
		break;
	    }
	}
    }

    close( td->fd );
    free( td->raddr );
    free( td );

    g_sco_threads--;
    _logf( LOG_INFO, "SCO threads: %d\n", g_sco_threads );

    pthread_exit( NULL );
    return NULL; /* make compiler happy */
}

int main( int argc, char *argv[] )
{
    int			ret;
    char		local_addr[18];
    char		buf[2048];
    char		*prog;
    int			features_present = 0, features = 0;
    struct pollfd 	ufds[2];
    pthread_t		dummy_thr;

    struct option long_options[] =
    {
	{"device",   required_argument, 0, 'd'},
	{"btdev",    required_argument, 0, 'b'},
	{"phone",    required_argument, 0, 'p'},
	{"nodaemon", no_argument,       0, 'n'},
	{"verbose",  no_argument,       0, 'v'},
	{"help",     no_argument,       0, 'h'},
	{0, 0, 0, 0}
    };

    prog = strrchr(argv[0],'/');
    if(!prog) prog = argv[0];
    else      prog++;

    conf_parse( HFP_CONFIG, params );

    while( (ret = getopt_long (argc, argv, "d:b:p:vnh", long_options, NULL)) != -1 )
    {
	switch(ret)
	{
    	    case 'd': strcpy( g_device, optarg ); break;
    	    case 'b': strcpy( g_btdev, optarg ); break;
	    case 'p': strcpy( g_phone, optarg ); break;
	    case 'v': g_verbose = 1; break;
	    case 'n': g_daemon = 0; break;
	    case 'h':
	    default:
		printf("Usage: %s [options]\n", prog);
		printf("Options:\n");
		printf("\t-d, --device sound_dev        Use device sound_dev\n");
		printf("\t-b, --btdev hciX              Use bluetooth adapter hciX\n");
		printf("\t-p, --phone bdaddr            Connect to phone with bdaddr\n");
		printf("\t-n, --nodaemon                Run in foreground\n");
		printf("\t-v, --verbose                 Verbose messages\n");
		printf("\t-h, --help                    Show this help\n");
		return 1;
	}
    }

    if( g_syslog ) openlog( prog, 0, LOG_USER );

    signal( SIGINT,  sig_handler );
    signal( SIGTERM, sig_handler );
    signal( SIGCHLD, SIG_IGN );

    if( g_daemon )
    { /* daemonize */
	if( !daemon_start() ) goto out_sound;
	if( !g_syslog )
	{ /* open syslog since we are daemon */
	    g_syslog = 1;
	}
    }

    if( g_syslog ) openlog( prog, 0, LOG_USER );

    ret = dsp_open( g_device, g_verbose );
    if( ret < 0 ) return 1;

    ret = hci_get_local( g_btdev, local_addr );
    if( ret < 0 ) return 1;

    _logf( LOG_INFO, "Using bluetooth adapter %s %s\n", g_btdev, local_addr );

    _logf( LOG_INFO, "Registering HandsFree Profile at SDP...\n" );

    g_sdp_hfp = sdp_register_hfp( hf_features, g_service_prov, g_service_desc );
    if( ! g_sdp_hfp )
    {
	_logf( LOG_ERR, "Can't register HFP to SDP!\n" );
	goto out_sound;
    }


reconnect:

    _logf( LOG_INFO, "Searching for the phone...\n" );

    g_channel = search_hfag( g_phone, &features_present, &features );
    if( g_channel < 0 )
    {
	_logf( g_persist ? LOG_INFO : LOG_ERR, "Can't find Handsfree Gateway at %s\n", g_phone );
	goto out_reconnect;
    }

    _logf( LOG_INFO, "Connecting to RFCOMM channel %d...\n", g_channel );

    g_rfcommfd = rfcomm_do_connect( g_phone, g_channel );
    if( g_rfcommfd < 0 ) goto out_reconnect;

    _logf( LOG_INFO, "Starting SCO server...\n" );
    g_scofd = sco_start_server( local_addr );
    if( g_scofd < 0 ) goto out_rfcomm;

    ufds[0].fd = g_rfcommfd;
    ufds[0].events = POLLIN;
    ufds[0].revents = 0;

    ufds[1].fd = g_scofd;
    ufds[1].events = POLLIN;
    ufds[1].revents = 0;

    while(1)
    {
	ret = poll( ufds, 2, -1 );

	if( ret < 0 )
	{
	    _logf( LOG_ERR, "poll() error: %s\n", strerror( ret ) );
	    break;
	}

	if( ufds[0].revents & POLLIN )
	{
	    ret = recv( g_rfcommfd, buf, sizeof(buf), 0 );
	    if( ret <= 0 )
	    {
		_logf( LOG_ERR, "recv error: %s\n", strerror( errno ) );
		break;
	    }
	    ret = rfcomm_engine_do( g_rfcommfd, buf, ret );
	    if( ret < 0 )
	    {
		break;
	    }
	}
	if( ufds[0].revents & ( POLLERR | POLLNVAL | POLLRDHUP ) )
	{ // error
	    _logf( LOG_ERR, "RFCOMM peer close connection\n" );
	    break;
	}

	if( ufds[1].revents & POLLIN )
	{
	    struct thread_data	*td;
	    struct sockaddr_sco	remote;
	    unsigned int	len = sizeof(struct sockaddr_sco);
	    char		addr[18];

	    ret = accept( g_scofd, (struct sockaddr*)&remote, &len );
	    if( ret < 0 )
	    {
		_logf( LOG_ERR, "accept error: %s\n", strerror( errno ) );
		break;
	    }

	    ba2str( &remote.sco_bdaddr, addr );
	    _logf( LOG_INFO, "Accepted SCO connection from %s\n", addr );

	    td = malloc( sizeof( struct thread_data ) );
	    if( td )
	    {
		memset( td, 0, sizeof( struct thread_data ) );
		td->fd = ret;
		td->raddr = strdup( addr );
		pthread_create( &dummy_thr, NULL, handle_connection, (void*)td );
	    }
	    else
	    {
		    _logf( LOG_ERR, "Failed to allocate thread data!" );
		    close( ret );
	    }
	}
	if( ufds[0].revents & ( POLLERR | POLLNVAL | POLLRDHUP ) )
	{ // error
	    _logf( LOG_ERR, "SCO server polling error\n" );
	    break;
	}
    }

    _logf( LOG_INFO, "Shutting down SCO server.\n" );
    close( g_scofd );

out_rfcomm:
    _logf( LOG_INFO, "Closing RFCOMM connection.\n" );
    close( g_rfcommfd );

out_reconnect:
    if( g_persist )
    {
	sleep( 5 );
	goto reconnect;
    }

    _logf( LOG_INFO, "Deregistering HandsFree Profile.\n" );
    sdp_deregister_hfp( g_sdp_hfp );

out_sound:
    dsp_close();

    if( g_syslog ) closelog();

    return 1;
}
