/*
 * Copyright (C) 2010 Andrew Kozin AKA Stanson <stanson@pangolin.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>

#include "log.h"
#include "rfcomm.h"
#include "sco.h"

extern int	g_verbose;
extern uint32_t	g_features;

extern int slc_established();

uint32_t	ag_features = 0;
uint32_t	hf_features = HF_FEATURE_CLI_PRESENTATION |
			      HF_FEATURE_REMOTE_VOLUME_CTRL ;

char		rfcomm_str[4096];
int		rfcomm_idx = 0;

#define STATE_DISCONNECTED	0
#define STATE_SENT_BRSF		1
#define STATE_SENT_CIND1	2
#define STATE_SENT_CIND2	3
#define STATE_SENT_CMER		4
#define STATE_SENT_CHLD		5
#define STATE_SENT_CMEE		6
#define STATE_SENT_CLIP		7
#define STATE_SENT_CCWA		8
#define STATE_SENT_CLCC		9
#define STATE_READY		10
int		rfcomm_state = STATE_DISCONNECTED;


int rfcomm_do_connect( char * addr, uint8_t channel )
{
    int			ret, sockfd;
    struct sockaddr_rc	to;

    sockfd = socket( AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM );
    if( sockfd < 0 )
    {
	_logf( LOG_ERR, "Can't create RFCOMM socket: %s\n", strerror( errno ) );
	return sockfd;
    }

    memset( &to, 0, sizeof( to ) );
    to.rc_family = AF_BLUETOOTH;
    to.rc_channel = channel;
    str2ba( addr, &to.rc_bdaddr );

    ret = connect( sockfd, (struct sockaddr*)&to, sizeof(to) );
    if( ret < 0 )
    {
	_logf( LOG_ERR, "Can't connect to %s RFCOMM channel %d\n", addr, channel );
	close( sockfd );
	return ret;
    }

    _logf( LOG_NOTICE, "Connected to %s RFCOMM channel %d\n", addr, channel );

    /* send our features to phone */
    ret = rfcomm_send_cmd( sockfd, "AT+BRSF=%d", hf_features );
    if( ret < 0 )
    {
	close( sockfd );
	return ret;
    }

    rfcomm_state = STATE_SENT_BRSF;

    return sockfd;
}

int rfcomm_send_cmd( int fd, const char * format, ... )
{
    va_list	arg;
    char	buf[256];
    int		ret;

    va_start( arg, format );
    vsnprintf( buf, sizeof(buf), format, arg );
    va_end( arg );

    _logf( LOG_INFO, "< %s\n", buf );

    strcat( buf, "\r" );

    ret = send( fd, buf, strlen(buf), 0 );
    if( ret < 0 )
    {
	_logf( LOG_ERR, "RFCOMM send() failed: %s\n", strerror( errno ) );
    }
    return ret;
}

int rfcomm_string_ready( int fd, char * s )
{
    int ret = 0, failed = 0;

    _logf( LOG_INFO, "> %s\n", s );

    if( !strcmp( s, "OK" ) || ( failed = !strncmp( s, "ERROR", 6) ) )
    { /* command completed */
	switch( rfcomm_state )
	{
	    case STATE_SENT_BRSF:
		ret = rfcomm_send_cmd( fd, "AT+CIND=?" );
		rfcomm_state = STATE_SENT_CIND1;
		break;
	    case STATE_SENT_CIND1:
		ret = rfcomm_send_cmd( fd, "AT+CIND?" );
		rfcomm_state = STATE_SENT_CIND2;
		break;
	    case STATE_SENT_CIND2:
		ret = rfcomm_send_cmd( fd, "AT+CMER=3,0,0,1" );
		rfcomm_state = STATE_SENT_CMER;
		break;
	    case STATE_SENT_CMER:
		if( ( hf_features & HF_FEATURE_THREE_WAY_CALLING ) &&
		    ( ag_features & AG_FEATURE_THREE_WAY_CALLING ) )
		{
		    ret = rfcomm_send_cmd( fd, "AT+CHLD=?" );
		    rfcomm_state = STATE_SENT_CHLD;
		    break;
		} // fall to next
	    case STATE_SENT_CHLD:
		// Service Level Connection established
		slc_established();
		if( ag_features & AG_FEATURE_EXTENDED_ERROR_CODES )
		{
		    ret = rfcomm_send_cmd( fd, "AT+CMEE=1" );
		    rfcomm_state = STATE_SENT_CMEE;
		    break;
		} // fall to next
	    case STATE_SENT_CMEE:
		ret = rfcomm_send_cmd( fd, "AT+CLIP=1" );
		rfcomm_state = STATE_SENT_CLIP;
		break;
	    case STATE_SENT_CLIP:
		ret = rfcomm_send_cmd( fd, "AT+CCWA=1" );
		rfcomm_state = STATE_SENT_CCWA;
		break;
	    case STATE_SENT_CCWA:
		ret = rfcomm_send_cmd( fd, "AT+CLCC" );
		rfcomm_state = STATE_SENT_CLCC;
		break;
	    case STATE_SENT_CLCC:
		rfcomm_state = STATE_READY;
		break;
	    default:
		break;
	}
    }
    else if( !strncmp( s, "RING", 4) )
    { /* incoming call */
    }
    else if( !strncmp( s, "+BRSF:", 6) )
    { /* get phone Bluetooth Retrieve Supported Feature */
	ag_features = strtoul( &s[6], NULL, 10 );
	if( g_verbose )
	{
	    _logf( LOG_INFO, "Phone Handsfree AG supported features:\n" );
	    if( ag_features & ( 1 << 0 ) ) _logf( LOG_INFO, "  Three way calling\n" );
	    if( ag_features & ( 1 << 1 ) ) _logf( LOG_INFO, "  EC / NR function\n" );
	    if( ag_features & ( 1 << 2 ) ) _logf( LOG_INFO, "  Voice recognition\n" );
	    if( ag_features & ( 1 << 3 ) ) _logf( LOG_INFO, "  In-band ring\n" );
	    if( ag_features & ( 1 << 4 ) ) _logf( LOG_INFO, "  Attach a number to voice tag\n" );
	    if( ag_features & ( 1 << 5 ) ) _logf( LOG_INFO, "  Ability to reject a call\n" );
	    if( ag_features & ( 1 << 6 ) ) _logf( LOG_INFO, "  Enhanced call status\n" );
	    if( ag_features & ( 1 << 7 ) ) _logf( LOG_INFO, "  Enhanced call control\n" );
	    if( ag_features & ( 1 << 8 ) ) _logf( LOG_INFO, "  Extended error result codes\n" );
	}
    }
    else if( !strncmp( s, "+VGS:", 5) )
    { /* Volume Gain Speaker */
    }
    else if( !strncmp( s, "+VGM:", 5) )
    { /* Volume Gain Microphone */
    }

    return ret;
}

int rfcomm_engine_do( int fd, char * s, int len )
{
    int		ret = 0;

    while( len )
    {
	switch( *s )
	{
	    case '\r':
		break;
	    case '\n':
		if( rfcomm_idx )
		{
		    rfcomm_str[rfcomm_idx] = 0;
		    rfcomm_idx = 0;
		    ret = rfcomm_string_ready( fd, rfcomm_str );
		    if( ret < 0 ) return ret;
		}
		break;
	    default:
		rfcomm_str[rfcomm_idx++] = *s;
		break;
	}
	s++;
	len--;
    }

    return ret;
}
