/*
 * Copyright (C) 2010 Andrew Kozin AKA Stanson <stanson@pangolin.ru>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "sdp.h"
#include "log.h"

sdp_session_t *sdp_register_hfp( uint16_t features, const char * prov, const char * desc )
{
    int			ret = 0;
    const char		*service_name = "Handsfree";
    uint8_t		rfcomm_channel = 3;
    uuid_t		root_uuid,
			l2cap_uuid,
			rfcomm_uuid,
			hfsc_uuid,
			gasc_uuid;
    sdp_list_t		*l2cap_list = 0,
        		*rfcomm_list = 0,
			*root_list = 0,
			*proto_list = 0;
    sdp_data_t		*channel = 0;
    sdp_profile_desc_t	hfp_pdesc;
    sdp_record_t	*record = sdp_record_alloc();
    sdp_session_t	*session = 0;


// Set the name, provider, and description

    sdp_set_info_attr( record, service_name, prov, desc );

// Make the service record publicly browsable

    // new list
    root_list = 0;

    sdp_uuid16_create( &root_uuid, PUBLIC_BROWSE_GROUP );
    root_list = sdp_list_append( root_list, &root_uuid );

    sdp_set_browse_groups( record, root_list );

    // free list
    sdp_list_free( root_list, 0 );

// Add class info to service record

    // new list
    root_list = 0;

    // add Handsfree class ID
    sdp_uuid16_create( &hfsc_uuid, HANDSFREE_SVCLASS_ID );
    root_list = sdp_list_append( root_list, &hfsc_uuid );

    // add Generic Audio class ID
    sdp_uuid16_create( &gasc_uuid, GENERIC_AUDIO_SVCLASS_ID );
    root_list = sdp_list_append( root_list, &gasc_uuid );

    // attach class info to service record
    sdp_set_service_classes( record, root_list );

    // free list
    sdp_list_free( root_list, 0 );

// Add protocol information to service record

    // new list
    root_list = 0;
    proto_list = 0;
    l2cap_list = 0;
    rfcomm_list = 0;

    // set l2cap information
    sdp_uuid16_create(&l2cap_uuid, L2CAP_UUID);
    l2cap_list = sdp_list_append( l2cap_list, &l2cap_uuid );

    // set rfcomm information
    sdp_uuid16_create( &rfcomm_uuid, RFCOMM_UUID );
    channel = sdp_data_alloc( SDP_UINT8, &rfcomm_channel );
    rfcomm_list = sdp_list_append( 0, &rfcomm_uuid );
    rfcomm_list = sdp_list_append( rfcomm_list, channel );

    // create protocol list
    proto_list = sdp_list_append( proto_list, l2cap_list );
    proto_list = sdp_list_append( proto_list, rfcomm_list );

    root_list = sdp_list_append( root_list, proto_list );

    // attach protocol information to service record
    sdp_set_access_protos( record, root_list );

    // free list
    sdp_data_free( channel );
    sdp_list_free( l2cap_list, 0 );
    sdp_list_free( rfcomm_list, 0 );
    sdp_list_free( proto_list, 0 );
    sdp_list_free( root_list, 0 );

// Add profile info

    // new list
    root_list = 0;

    sdp_uuid16_create( &hfp_pdesc.uuid, HANDSFREE_PROFILE_ID );
    hfp_pdesc.version = 0x0105;
    root_list = sdp_list_append( root_list, &hfp_pdesc );

    // attach profile information to service record
    sdp_set_profile_descs( record, root_list );

    // free list
    sdp_list_free( root_list, 0 );

// Add caps

    features &= 0x1F;
    sdp_attr_add_new( record, SDP_ATTR_SUPPORTED_FEATURES, SDP_UINT16, &features );

// connect to the local SDP server and register the service record

    session = sdp_connect( BDADDR_ANY, BDADDR_LOCAL, SDP_RETRY_IF_BUSY );
    if( session )
    {
        ret = sdp_record_register( session, record, 0 );
        if( ret == 0 )
        {
    	    _logf( LOG_INFO, "HandsFree Profile registered.\n" );
        }
    }
    return session;
}

int sdp_deregister_hfp( sdp_session_t * session )
{
    return sdp_close( session );
}

int search_hfag( char * addr, int * features_present, int * features )
{
    bdaddr_t		target;
    sdp_session_t	*session = 0;
    sdp_list_t		*r,
			*response_list = 0,
			*search_list = 0,
			*attr_list = 0;
    int			ret = -1, brsf = 0;
    uuid_t		svclass;
    uint16_t		attr1, attr2;

    str2ba( addr, &target );

    _logf( LOG_INFO, "Connecting to phone %s\n", addr );
    // connect to the SDP server running on the remote machine
    session = sdp_connect( BDADDR_ANY, &target, SDP_RETRY_IF_BUSY );
    if( !session )
    {
        _logf( LOG_ERR, "SDP connect to %s failed!", addr );
        goto out;
    }

    // Search for class
    sdp_uuid16_create( &svclass, HANDSFREE_AGW_SVCLASS_ID );
    search_list = sdp_list_append( search_list, &svclass );

    attr1 = SDP_ATTR_PROTO_DESC_LIST;
    attr_list = sdp_list_append( attr_list, &attr1 );
    attr2 = SDP_ATTR_SUPPORTED_FEATURES;
    attr_list = sdp_list_append( attr_list, &attr2 );

    _logf( LOG_INFO, "Searching SDP records for Handsfree Audio Gateway...\n" );

    // get a list of service records that have UUID 0xabcd
    ret = sdp_service_search_attr_req( session, search_list, 
		SDP_ATTR_REQ_INDIVIDUAL, attr_list, &response_list);

    if( ret < 0 )
    {
        _logf( LOG_ERR, "SDP search on %s failed!", addr );
        goto out_close;
    }

    r = response_list;

    // go through each of the service records
    for( ; r; r = r->next )
    {
        sdp_record_t	*rec = (sdp_record_t*) r->data;
        sdp_list_t	*proto_list;

	ret = sdp_get_access_protos( rec, &proto_list );
        if( ret == 0 )
        {
	    ret = sdp_get_proto_port( proto_list, RFCOMM_UUID );
	    if( ret > 0 )
	    {
    		_logf( LOG_INFO, "Found RFCOMM channel %d\n", ret );

		*features_present = 0;
		*features = 0;

		/* See if there's a capabilities record */
		if( ! sdp_get_int_attr( rec, SDP_ATTR_SUPPORTED_FEATURES, &brsf ) )
		{
		    *features_present = 1;
		    *features = brsf;
    		    _logf( LOG_INFO, "Found features 0x%08X\n", brsf );
		}
	    }
	    else
	    {
		ret = -1;
	    }
    	    sdp_list_free( proto_list, 0 );
        }
        sdp_record_free( rec );
    }
out_close:
    sdp_list_free( attr_list, 0 );
    sdp_list_free( search_list, 0 );
    sdp_close( session );
out:
    return ret;
}
